@@ -402,9 +402,9 @@ class MultiMapping(typing.Generic[KT, VT], typing.Mapping[KT, VT]):
         return list(self._list)
 
     def __eq__(self, other: typing.Any) -> bool:
-        if not isinstance(other, self.__class__):
-            return False
-        return sorted(self._list) == sorted(other._list)
+        return isinstance(other, type(self)) and sorted(self._list) == sorted(
+            other._list
+        )
 
     def __repr__(self) -> str:
         class_name = self.__class__.__name__
@@ -539,17 +539,20 @@ class MutableHeaders(Headers, typing.MutableMapping[str, str]):
     __slots__ = Headers.__slots__
 
     def __setitem__(self, key: str, value: str) -> None:
+        if "\n" in key or "\r" in key or "\0" in key:
+            raise ValueError("Header names must not contain control characters.")
+        if "\n" in value or "\r" in value or "\0" in value:
+            raise ValueError("Header values must not contain control characters.")
         self._dict[key.lower()] = value
 
     def __delitem__(self, key: str) -> None:
         del self._dict[key.lower()]
 
     def append(self, key: str, value: str) -> None:
-        key = key.lower()
-        if key in self._dict:
-            self._dict[key] = f"{self._dict[key]}, {value}"
+        if key in self:
+            self[key] = f"{self[key]}, {value}"
         else:
-            self._dict[key] = value
+            self[key] = value
 
 
 class UploadFile:
