@@ -1,4 +1,5 @@
 import asyncio
+import json
 import tempfile
 from functools import partial
 from inspect import cleandoc
@@ -28,15 +29,19 @@ from baize.asgi import (
     Subpaths,
     WebSocket,
     WebSocketDisconnect,
-    middleware,
+    decorator,
     request_response,
     websocket_session,
+    middleware,
+    NextRequest,
+    NextResponse,
 )
-from baize.datastructures import UploadFile
+from baize.datastructures import FormData, UploadFile
 from baize.exceptions import (
     HTTPException,
     MalformedJSON,
     MalformedMultipart,
+    RequestEntityTooLarge,
     UnsupportedMediaType,
 )
 from baize.typing import Message, ServerSentEvent
@@ -76,7 +81,9 @@ async def test_request_url():
         response = JSONResponse(data)
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/123?a=abc")
         assert response.json() == {
             "method": "GET",
@@ -95,7 +102,9 @@ async def test_request_query_params():
         response = JSONResponse({"params": params})
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/?a=123&b=456")
         assert response.json() == {"params": {"a": "123", "b": "456"}}
 
@@ -109,7 +118,9 @@ async def test_request_headers():
         response = JSONResponse({"headers": headers})
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/", headers={"host": "example.org"})
         assert response.json() == {
             "headers": {
@@ -130,7 +141,9 @@ async def test_request_client():
         )
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/")
         assert response.json() == {"host": "127.0.0.1", "port": 123}
 
@@ -143,12 +156,15 @@ async def test_request_body():
         response = JSONResponse({"body": body.decode()})
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/")
         assert response.json() == {"body": ""}
 
         response = await client.post("/", json={"a": "123"})
-        assert response.json() == {"body": '{"a": "123"}'}
+        inner_json = json.loads(response.json()["body"])
+        assert inner_json == {"a": "123"}
 
         response = await client.post("/", content="abc")
         assert response.json() == {"body": "abc"}
@@ -164,12 +180,15 @@ async def test_request_stream():
         response = JSONResponse({"body": body.decode()})
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/")
         assert response.json() == {"body": ""}
 
         response = await client.post("/", json={"a": "123"})
-        assert response.json() == {"body": '{"a": "123"}'}
+        inner_json = json.loads(response.json()["body"])
+        assert inner_json == {"a": "123"}
 
         response = await client.post("/", content="abc")
         assert response.json() == {"body": "abc"}
@@ -184,7 +203,9 @@ async def test_request_form_urlencoded():
         await response(scope, receive, send)
         await request.close()
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.post("/", data={"abc": "123 @"})
         assert response.json() == {"form": {"abc": "123 @"}}
 
@@ -206,7 +227,9 @@ async def test_request_multipart_form():
         await response(scope, receive, send)
         await request.close()
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         with tempfile.SpooledTemporaryFile(1024) as file:
             file.write(b"temporary file")
             file.seek(0, 0)
@@ -221,6 +244,46 @@ async def test_request_multipart_form():
             )
 
 
+@pytest.mark.asyncio
+async def test_request_multipart_form_limit():
+    class FRequest(Request):
+        async def _parse_multipart(self, boundary: bytes, charset: str) -> FormData:
+            from baize.multipart_helper import parse_async_stream as parse_multipart
+
+            return FormData(
+                await parse_multipart(
+                    self.stream(),
+                    boundary,
+                    charset,
+                    file_factory=UploadFile,
+                    max_form_parts=2,
+                    max_form_memory_size=1024,
+                )
+            )
+
+    async def app(scope, receive, send):
+        await FRequest(scope, receive).form
+
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
+        with pytest.raises(RequestEntityTooLarge):
+            with tempfile.SpooledTemporaryFile(1024) as file:
+                file.write(b"temporary file")
+                file.seek(0, 0)
+                await client.post(
+                    "/", data={"part1": "1", "part2": "2"}, files={"file-key": file}
+                )
+
+        with pytest.raises(RequestEntityTooLarge):
+            with tempfile.SpooledTemporaryFile(1024) as file:
+                file.write(b"temporary file")
+                file.seek(0, 0)
+                await client.post(
+                    "/", data={"abc": "*" * 2048}, files={"file-key": file}
+                )
+
+
 @pytest.mark.asyncio
 async def test_request_body_then_stream():
     async def app(scope, receive, send):
@@ -232,7 +295,9 @@ async def test_request_body_then_stream():
         response = JSONResponse({"body": body.decode(), "stream": chunks.decode()})
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.post("/", content="abc")
         assert response.json() == {"body": "abc", "stream": "abc"}
 
@@ -251,7 +316,9 @@ async def test_request_stream_then_body():
         response = JSONResponse({"body": body.decode(), "stream": chunks.decode()})
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.post("/", content="abc")
         assert response.json() == {"body": "<stream consumed>", "stream": "abc"}
 
@@ -264,7 +331,9 @@ async def test_request_json():
         response = JSONResponse({"json": data})
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.post("/", json={"a": "123"})
         assert response.json() == {"json": {"a": "123"}}
 
@@ -297,7 +366,9 @@ async def test_request_without_setting_receive():
         response = JSONResponse({"json": data})
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.post("/", json={"a": "123"})
         assert response.json() == {"json": "Receive channel not available"}
 
@@ -339,7 +410,9 @@ async def test_request_is_disconnected():
         await response(scope, receive, send)
         disconnected_after_response = await request.is_disconnected()
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/")
         assert response.json() == {"disconnected": False}
         assert disconnected_after_response
@@ -358,7 +431,9 @@ async def test_request_cookies():
 
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/")
         assert response.text == "Hello, world!"
         response = await client.get("/")
@@ -393,7 +468,9 @@ async def test_cookie_lenient_parsing():
         response = JSONResponse({"cookies": request.cookies})
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/", headers={"cookie": tough_cookie})
         result = response.json()
         assert len(result["cookies"]) == 4
@@ -429,7 +506,9 @@ async def test_cookies_edge_cases(set_cookie, expected):
         response = JSONResponse({"cookies": request.cookies})
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/", headers={"cookie": set_cookie})
         result = response.json()
         assert result["cookies"] == expected
@@ -470,7 +549,9 @@ async def test_cookies_invalid(set_cookie, expected):
         response = JSONResponse({"cookies": request.cookies})
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/", headers={"cookie": set_cookie})
         result = response.json()
         assert result["cookies"] == expected
@@ -489,7 +570,9 @@ async def test_response_headers():
         response.headers["x-header-2"] = "789"
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/")
         assert response.headers["x-header-1"] == "123"
         assert response.headers["x-header-2"] == "789"
@@ -510,7 +593,9 @@ async def test_set_cookie():
         samesite="none",
     )
 
-    async with httpx.AsyncClient(app=response, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(response)
+    ) as client:
         response = await client.get("/")
         assert response.text == "Hello, world!"
 
@@ -526,7 +611,9 @@ async def test_delete_cookie():
             response.set_cookie("mycookie", "myvalue")
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/")
         assert response.cookies["mycookie"]
         response = await client.get("/")
@@ -542,7 +629,9 @@ async def test_redirect_response():
             response = RedirectResponse("/")
         await response(scope, receive, send)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/redirect", follow_redirects=True)
         assert response.text == "hello, world"
         assert response.url == "http://testserver/"
@@ -555,7 +644,8 @@ async def test_stream_response():
             yield str(i).encode("utf-8")
 
     async with httpx.AsyncClient(
-        app=StreamResponse(generator(10)), base_url="http://testServer/"
+        transport=httpx.ASGITransport(StreamResponse(generator(10))),
+        base_url="http://testServer/",
     ) as client:
         response = await client.get("/")
         assert response.content == b"".join(str(i).encode("utf-8") for i in range(10))
@@ -588,7 +678,9 @@ async def test_file_response(tmp_path: Path, response_class: Type[FileResponse])
     async def app(scope, r, s):
         return await response_class(str(filepath))(scope, r, s)
 
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         response = await client.get("/")
         assert response.status_code == 200
         assert response.headers["content-length"] == str(len(README.encode("utf8")))
@@ -639,12 +731,12 @@ async def test_file_response(tmp_path: Path, response_class: Type[FileResponse])
         assert response.status_code == 400
 
         response = await client.head(
-            "/", headers={"Range": f"bytes=0-{len(README.encode('utf8'))+1}"}
+            "/", headers={"Range": f"bytes=0-{len(README.encode('utf8')) + 1}"}
         )
         assert response.status_code == 206
 
         response = await client.head(
-            "/", headers={"Range": f"bytes={len(README.encode('utf8'))+1}-"}
+            "/", headers={"Range": f"bytes={len(README.encode('utf8')) + 1}-"}
         )
         assert response.status_code == 416
         assert response.headers["Content-Range"] == f"*/{len(README.encode('utf8'))}"
@@ -662,7 +754,7 @@ async def test_file_response_with_download_name(tmp_path: Path):
     filepath.write_bytes(README.encode("utf8"))
     file_response = FileResponse(str(filepath), download_name="README.txt")
     async with httpx.AsyncClient(
-        app=file_response, base_url="http://testServer/"
+        transport=httpx.ASGITransport(file_response), base_url="http://testServer/"
     ) as client:
         response = await client.get("/")
         assert (
@@ -695,21 +787,25 @@ async def test_send_event_response():
     )
 
     async with httpx.AsyncClient(
-        app=SendEventResponse(send_events(), ping_interval=0.1),
+        transport=httpx.ASGITransport(
+            SendEventResponse(send_events(), ping_interval=0.1)
+        ),
         base_url="http://testServer/",
     ) as client:
         async with client.stream("GET", "/") as resp:
             resp.raise_for_status()
             events = ""
             async for line in resp.aiter_lines():
-                events += line
+                events += line + "\n"
             assert events.replace(": ping\n\n", "") == expected_events
 
     async with httpx.AsyncClient(
-        app=SendEventResponse(
-            send_events(),
-            headers={"custom-header": "value"},
-            ping_interval=0.1,
+        transport=httpx.ASGITransport(
+            SendEventResponse(
+                send_events(),
+                headers={"custom-header": "value"},
+                ping_interval=0.1,
+            )
         ),
         base_url="http://testServer/",
     ) as client:
@@ -718,10 +814,57 @@ async def test_send_event_response():
             assert resp.headers["custom-header"] == "value"
             events = ""
             async for line in resp.aiter_lines():
-                events += line
+                events += line + "\n"
             assert events.replace(": ping\n\n", "") == expected_events
 
 
+@pytest.mark.asyncio
+async def test_send_event_response_raise_exception():
+    async def send_events() -> AsyncGenerator[ServerSentEvent, None]:
+        yield ServerSentEvent(data="hello\nworld")
+        await asyncio.sleep(0.2)
+        raise Exception("Something went wrong")
+
+    async with httpx.AsyncClient(
+        transport=httpx.ASGITransport(
+            SendEventResponse(send_events(), ping_interval=0.1)
+        ),
+        base_url="http://testServer/",
+    ) as client:
+        with pytest.raises(Exception, match="Something went wrong"):
+            async with client.stream("GET", "/") as resp:
+                resp.raise_for_status()
+                events = ""
+                async for line in resp.aiter_lines():
+                    events += line
+
+
+@pytest.mark.asyncio
+async def test_send_event_response_be_killed():
+    killed = False
+
+    async def send_events() -> AsyncGenerator[ServerSentEvent, None]:
+        nonlocal killed
+
+        try:
+            for _ in range(10):
+                yield ServerSentEvent(data="hello\nworld")
+        finally:
+            killed = True
+
+    async with httpx.AsyncClient(
+        transport=httpx.ASGITransport(
+            SendEventResponse(send_events(), ping_interval=0.1)
+        ),
+        base_url="http://testServer/",
+    ) as client:
+        async with client.stream("GET", "/") as resp:
+            resp.raise_for_status()
+            await resp.aread()
+
+    assert killed
+
+
 @pytest.mark.parametrize(
     "response_class",
     [
@@ -844,6 +987,7 @@ def test_client_close():
             await websocket.receive_text()
         except WebSocketDisconnect as exc:
             assert exc.code == 1001
+            assert repr(exc) == "WebSocketDisconnect(code=1001, reason=)"
 
     client = TestClient(app)
     with client.websocket_connect("/") as websocket:
@@ -883,7 +1027,8 @@ def test_rejected_connection():
 
     client = TestClient(app)
     with pytest.raises(WebSocketDisconnect) as exc:
-        client.websocket_connect("/")
+        with client.websocket_connect("/"):
+            pass
     assert exc.value.code == 1001
 
 
@@ -922,8 +1067,7 @@ def test_websocket_scope_interface():
     async def mock_receive() -> Message:
         return {}
 
-    async def mock_send(message: Message) -> None:
-        ...  # pragma: no cover
+    async def mock_send(message: Message) -> None: ...
 
     websocket = WebSocket(
         {"type": "websocket", "path": "/abc/", "headers": []},
@@ -940,53 +1084,6 @@ def test_websocket_scope_interface():
 # ######################################################################################
 
 
-@pytest.mark.asyncio
-async def test_request_response():
-    @request_response
-    async def view(request: Request) -> Response:
-        return PlainTextResponse(await request.body)
-
-    async with httpx.AsyncClient(app=view, base_url="http://testServer/") as client:
-        assert (await client.get("/")).text == ""
-        assert (await client.post("/", content="hello")).text == "hello"
-
-    client = TestClient(view)
-    with pytest.raises(WebSocketDisconnect) as exc:
-        with client.websocket_connect("/"):
-            pass
-    assert exc.value.code == 1000
-
-
-@pytest.mark.asyncio
-async def test_websocket_session():
-    @websocket_session
-    async def view(websocket: WebSocket) -> None:
-        await websocket.accept()
-        await websocket.close()
-
-    async with httpx.AsyncClient(app=view, base_url="http://testServer/") as client:
-        assert (await client.get("/")).status_code == 404
-
-
-@pytest.mark.asyncio
-async def test_middleware():
-    @middleware
-    async def middleware_func(
-        request: Request, handler: Callable[[Request], Awaitable[Response]]
-    ) -> Response:
-        response = await handler(request)
-        response.headers["X-Middleware"] = "1"
-        return response
-
-    @request_response
-    @middleware_func
-    async def view(request: Request) -> Response:
-        return PlainTextResponse(await request.body)
-
-    async with httpx.AsyncClient(app=view, base_url="http://testServer/") as client:
-        assert (await client.get("/")).headers["X-Middleware"] == "1"
-
-
 @pytest.mark.asyncio
 async def test_router():
     @request_response
@@ -1002,11 +1099,13 @@ async def test_router():
         ("/redirect", redirect),
         ("/{path}", path),
     )
-    async with httpx.AsyncClient(app=router, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(router)
+    ) as client:
         assert (await client.get("/")).text == "homepage"
         assert (await client.get("/baize")).json() == {"path": "baize"}
         assert (await client.get("/baize/")).status_code == 404
-        assert (await (client.get("/redirect"))).headers["location"] == "/cat"
+        assert (await client.get("/redirect")).headers["location"] == "/cat"
 
 
 @pytest.mark.asyncio
@@ -1020,9 +1119,11 @@ async def test_subpaths():
         return PlainTextResponse(request["path"])
 
     async with httpx.AsyncClient(
-        app=Subpaths(
-            ("/frist", root),
-            ("/latest", path),
+        transport=httpx.ASGITransport(
+            Subpaths(
+                ("/frist", root),
+                ("/latest", path),
+            )
         ),
         base_url="http://testServer/",
     ) as client:
@@ -1031,9 +1132,11 @@ async def test_subpaths():
         assert (await client.get("/latest")).text == ""
 
     async with httpx.AsyncClient(
-        app=Subpaths(
-            ("", path),
-            ("/root", root),
+        transport=httpx.ASGITransport(
+            Subpaths(
+                ("", path),
+                ("/root", root),
+            )
         ),
         base_url="http://testServer/",
     ) as client:
@@ -1044,9 +1147,11 @@ async def test_subpaths():
 @pytest.mark.asyncio
 async def test_hosts():
     async with httpx.AsyncClient(
-        app=Hosts(
-            ("testServer", PlainTextResponse("testServer")),
-            (".*", PlainTextResponse("default host")),
+        transport=httpx.ASGITransport(
+            Hosts(
+                ("testServer", PlainTextResponse("testServer")),
+                (".*", PlainTextResponse("default host")),
+            )
         ),
         base_url="http://testServer/",
     ) as client:
@@ -1067,10 +1172,13 @@ async def test_hosts():
     [
         Files(Path(__file__).absolute().parent.parent / "baize"),
         Files(".", "baize"),
+        Files(".", "baize", handle_404=PlainTextResponse("", 404)),
     ],
 )
 async def test_files(app):
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         resp = await client.get("/py.typed")
         assert resp.text == ""
 
@@ -1107,11 +1215,15 @@ async def test_files(app):
             )
         ).status_code == 304
 
-        with pytest.raises(HTTPException):
-            await client.get("/")
+        if app.handle_404 is None:
+            with pytest.raises(HTTPException):
+                await client.get("/")
 
-        with pytest.raises(HTTPException):
-            await client.get("/%2E%2E/baize/%2E%2E/%2E%2E/README.md")
+            with pytest.raises(HTTPException):
+                await client.get("/%2E%2E/baize/%2E%2E/%2E%2E/README.md")
+
+        else:
+            assert (await client.get("/")).status_code == 404
 
 
 @pytest.mark.asyncio
@@ -1125,7 +1237,9 @@ async def test_pages(tmpdir):
     )
 
     app = Pages(tmpdir)
-    async with httpx.AsyncClient(app=app, base_url="http://testServer/") as client:
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
         resp = await client.get("/")
         assert resp.status_code == 200
         assert resp.text == "<html><body>index</body></html>"
@@ -1160,3 +1274,118 @@ async def test_pages(tmpdir):
 
         with pytest.raises(HTTPException):
             await client.get("/d")
+
+    app = Pages(tmpdir, handle_404=PlainTextResponse("", 404))
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
+        assert (await client.get("/d")).status_code == 404
+
+
+# ######################################################################################
+# #################################### Shortcut tests ##################################
+# ######################################################################################
+
+
+@pytest.mark.asyncio
+async def test_request_response():
+    @request_response
+    async def view(request: Request) -> Response:
+        return PlainTextResponse(await request.body)
+
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(view)
+    ) as client:
+        assert (await client.get("/")).text == ""
+        assert (await client.post("/", content="hello")).text == "hello"
+
+    client = TestClient(view)
+    with pytest.raises(
+        (WebSocketDisconnect, starlette.testclient.WebSocketDenialResponse)
+    ):
+        with client.websocket_connect("/"):
+            pass
+
+
+@pytest.mark.asyncio
+async def test_websocket_session():
+    @websocket_session
+    async def view(websocket: WebSocket) -> None:
+        await websocket.accept()
+        await websocket.close()
+
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(view)
+    ) as client:
+        assert (await client.get("/")).status_code == 404
+
+
+@pytest.mark.asyncio
+async def test_decorator():
+    @decorator
+    async def decorator_func(
+        request: Request, handler: Callable[[Request], Awaitable[Response]]
+    ) -> Response:
+        response = await handler(request)
+        response.headers["X-Middleware"] = "1"
+        return response
+
+    @request_response
+    @decorator_func
+    async def view(request: Request) -> Response:
+        return PlainTextResponse(await request.body)
+
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(view)
+    ) as client:
+        assert (await client.get("/")).headers["X-Middleware"] == "1"
+
+
+@pytest.mark.asyncio
+async def test_middleware():
+    @middleware
+    async def m(
+        request: NextRequest, handler: Callable[[NextRequest], Awaitable[NextResponse]]
+    ) -> Response:
+        response = await handler(request)
+        response.headers["X-Middleware"] = "1"
+        return response
+
+    @m
+    @request_response
+    async def view(request: Request) -> Response:
+        return PlainTextResponse(await request.body)
+
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(view)
+    ) as client:
+        response = await client.post("/", content="x")
+        assert response.headers["X-Middleware"] == "1"
+        assert response.text == "x"
+
+
+@pytest.mark.asyncio
+async def test_next_request():
+    async def app(scope, receive, send):
+        request = NextRequest(scope, receive)
+        request["KEY"] = "VALUE"
+        del request["KEY"]
+        response = await NextResponse.from_app(PlainTextResponse(""), request)
+        await response(scope, receive, send)
+
+    async with httpx.AsyncClient(
+        base_url="http://testServer/", transport=httpx.ASGITransport(app)
+    ) as client:
+        assert (await client.get("/")).status_code == 200
+
+    async def read_body(scope, receive, send):
+        request = NextRequest(scope, receive)
+        await request.body
+        response = await NextResponse.from_app(PlainTextResponse(""), request)
+        await response(scope, receive, send)
+
+    async with httpx.AsyncClient(
+        transport=httpx.ASGITransport(read_body), base_url="http://testServer/"
+    ) as client:
+        with pytest.raises(RuntimeError):
+            await client.post("/", content="hello")
