@@ -1,3 +1,4 @@
+import json
 import tempfile
 import time
 from inspect import cleandoc
@@ -7,11 +8,12 @@ from typing import Callable, Generator
 import httpx
 import pytest
 
-from baize.datastructures import Address, UploadFile
+from baize.datastructures import Address, FormData, UploadFile
 from baize.exceptions import (
     HTTPException,
     MalformedJSON,
     MalformedMultipart,
+    RequestEntityTooLarge,
     UnsupportedMediaType,
 )
 from baize.typing import ServerSentEvent
@@ -30,7 +32,10 @@ from baize.wsgi import (
     SendEventResponse,
     StreamResponse,
     Subpaths,
+    decorator,
     middleware,
+    NextRequest,
+    NextResponse,
     request_response,
 )
 
@@ -58,7 +63,9 @@ def test_request_url():
         response = PlainTextResponse(request.method + " " + str(request.url))
         return response(environ, start_response)
 
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         response = client.get("/123?a=abc")
         assert response.text == "GET http://testserver/123?a=abc"
 
@@ -73,7 +80,9 @@ def test_request_query_params():
         response = JSONResponse({"params": params})
         return response(environ, start_response)
 
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         response = client.get("/?a=123&b=456")
         assert response.json() == {"params": {"a": "123", "b": "456"}}
 
@@ -86,7 +95,9 @@ def test_request_headers():
         response = JSONResponse({"headers": headers})
         return response(environ, start_response)
 
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         response = client.get("/", headers={"host": "example.org"})
         assert response.json() == {
             "headers": {
@@ -106,7 +117,9 @@ def test_request_client():
         )
         return response(environ, start_response)
 
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         response = client.get("/")
         assert response.json() == {"host": None, "port": None}
 
@@ -121,12 +134,15 @@ def test_request_body():
         response = JSONResponse({"body": body.decode()})
         return response(environ, start_response)
 
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         response = client.get("/")
         assert response.json() == {"body": ""}
 
         response = client.post("/", json={"a": "123"})
-        assert response.json() == {"body": '{"a": "123"}'}
+        inner_json = json.loads(response.json()["body"])
+        assert inner_json == {"a": "123"}
 
         response = client.post("/", content="abc")
         assert response.json() == {"body": "abc"}
@@ -141,12 +157,14 @@ def test_request_stream():
         response = PlainTextResponse(body)
         return response(environ, start_response)
 
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         response = client.get("/")
         assert response.text == ""
 
         response = client.post("/", json={"a": "123"})
-        assert response.text == '{"a": "123"}'
+        assert response.json() == {"a": "123"}
 
         response = client.post("/", content="abc")
         assert response.text == "abc"
@@ -160,7 +178,9 @@ def test_request_form_urlencoded():
         return response(environ, start_response)
         request.close()
 
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         response = client.post("/", data={"abc": "123 @"})
         assert response.json() == {"form": {"abc": "123 @"}}
 
@@ -181,7 +201,9 @@ def test_request_multipart_form():
         request.close()
         return response(environ, start_response)
 
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         with tempfile.SpooledTemporaryFile(1024) as file:
             file.write(b"temporary file")
             file.seek(0, 0)
@@ -194,6 +216,43 @@ def test_request_multipart_form():
             )
 
 
+def test_request_multipart_form_limit():
+    class FRequest(Request):
+        def _parse_multipart(self, boundary: bytes, charset: str) -> FormData:
+            from baize.multipart_helper import parse_stream as parse_multipart
+
+            return FormData(
+                parse_multipart(
+                    self.stream(),
+                    boundary,
+                    charset,
+                    file_factory=UploadFile,
+                    max_form_parts=2,
+                    max_form_memory_size=1024,
+                )
+            )
+
+    def app(environ, start_response):
+        FRequest(environ).form
+
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
+        with pytest.raises(RequestEntityTooLarge):
+            with tempfile.SpooledTemporaryFile(1024) as file:
+                file.write(b"temporary file")
+                file.seek(0, 0)
+                client.post(
+                    "/", data={"part1": "1", "part2": "2"}, files={"file-key": file}
+                )
+
+        with pytest.raises(RequestEntityTooLarge):
+            with tempfile.SpooledTemporaryFile(1024) as file:
+                file.write(b"temporary file")
+                file.seek(0, 0)
+                client.post("/", data={"abc": "*" * 2048}, files={"file-key": file})
+
+
 def test_request_body_then_stream():
     def app(environ, start_response):
         request = Request(environ)
@@ -204,7 +263,9 @@ def test_request_body_then_stream():
         response = JSONResponse({"body": body.decode(), "stream": chunks.decode()})
         return response(environ, start_response)
 
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         response = client.post("/", content="abc")
         assert response.json() == {"body": "abc", "stream": "abc"}
 
@@ -222,7 +283,9 @@ def test_request_stream_then_body():
         response = JSONResponse({"body": body.decode(), "stream": chunks.decode()})
         return response(environ, start_response)
 
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         response = client.post("/", content="abc")
         assert response.json() == {"body": "<stream consumed>", "stream": "abc"}
 
@@ -234,7 +297,9 @@ def test_request_json():
         response = JSONResponse({"json": data})
         return response(environ, start_response)
 
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         response = client.post("/", json={"a": "123"})
         assert response.json() == {"json": {"a": "123"}}
 
@@ -262,7 +327,9 @@ def test_request_accpet():
             response = PlainTextResponse(data)
         return response(environ, start_response)
 
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         response = client.get("/", headers={"Accept": "application/json"})
         assert response.json() == {"data": data}
 
@@ -273,7 +340,9 @@ def test_request_accpet():
 
 
 def test_unknown_status():
-    with httpx.Client(app=Response(600), base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(Response(600))
+    ) as client:
         response = client.get("/")
         assert response.status_code == 600
 
@@ -286,7 +355,9 @@ def test_redirect_response():
             response = RedirectResponse("/")
         return response(environ, start_response)
 
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         response = client.get("/redirect", follow_redirects=True)
         assert response.text == "hello, world"
         assert response.url == "http://testserver/"
@@ -298,7 +369,8 @@ def test_stream_response():
             yield str(i).encode("utf-8")
 
     with httpx.Client(
-        app=StreamResponse(generator(10)), base_url="http://testServer/"
+        transport=httpx.WSGITransport(StreamResponse(generator(10))),
+        base_url="http://testServer/",
     ) as client:
         response = client.get("/")
         assert response.content == b"".join(str(i).encode("utf-8") for i in range(10))
@@ -319,7 +391,9 @@ def test_file_response(tmp_path: Path):
     filepath = tmp_path / "README.txt"
     filepath.write_bytes(README.encode("utf8"))
     file_response = FileResponse(str(filepath))
-    with httpx.Client(app=file_response, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(file_response)
+    ) as client:
         response = client.get("/")
         assert response.status_code == 200
         assert response.headers["content-length"] == str(len(README.encode("utf8")))
@@ -372,7 +446,7 @@ def test_file_response(tmp_path: Path):
         response = client.head(
             "/",
             headers={
-                "Range": f"bytes={len(README.encode('utf8'))+1}-{len(README.encode('utf8'))+12}"
+                "Range": f"bytes={len(README.encode('utf8')) + 1}-{len(README.encode('utf8')) + 12}"
             },
         )
         assert response.status_code == 416
@@ -388,7 +462,9 @@ def test_file_response_with_download_name(tmp_path: Path):
     filepath = tmp_path / "README"
     filepath.write_bytes(README.encode("utf8"))
     file_response = FileResponse(str(filepath), download_name="README.txt")
-    with httpx.Client(app=file_response, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(file_response)
+    ) as client:
         response = client.get("/")
         assert (
             response.headers["content-disposition"]
@@ -419,21 +495,25 @@ def test_send_event_response():
     )
 
     with httpx.Client(
-        app=SendEventResponse(send_events(), ping_interval=0.1),
+        transport=httpx.WSGITransport(
+            SendEventResponse(send_events(), ping_interval=0.1)
+        ),
         base_url="http://testServer/",
     ) as client:
         with client.stream("GET", "/") as resp:
             resp.raise_for_status()
             events = ""
             for line in resp.iter_lines():
-                events += line
+                events += line + "\n"
             assert events.replace(": ping\n\n", "") == expected_events
 
     with httpx.Client(
-        app=SendEventResponse(
-            send_events(),
-            headers={"custom-header": "value"},
-            ping_interval=0.1,
+        transport=httpx.WSGITransport(
+            SendEventResponse(
+                send_events(),
+                headers={"custom-header": "value"},
+                ping_interval=0.1,
+            )
         ),
         base_url="http://testServer/",
     ) as client:
@@ -442,10 +522,55 @@ def test_send_event_response():
             assert resp.headers["custom-header"] == "value"
             events = ""
             for line in resp.iter_lines():
-                events += line
+                events += line + "\n"
             assert events.replace(": ping\n\n", "") == expected_events
 
 
+def test_send_event_response_raise_exception():
+    def send_events() -> Generator[ServerSentEvent, None, None]:
+        yield ServerSentEvent(data="hello\nworld")
+        time.sleep(0.2)
+        raise Exception("Something went wrong")
+
+    with httpx.Client(
+        transport=httpx.WSGITransport(
+            SendEventResponse(send_events(), ping_interval=0.1)
+        ),
+        base_url="http://testServer/",
+    ) as client:
+        with client.stream("GET", "/") as resp:
+            resp.raise_for_status()
+            events = ""
+            with pytest.raises(Exception, match="Something went wrong"):
+                for line in resp.iter_lines():
+                    events += line
+
+
+def test_send_event_response_be_killed():
+    killed = False
+
+    def send_events() -> Generator[ServerSentEvent, None, None]:
+        nonlocal killed
+
+        try:
+            for _ in range(10):
+                yield ServerSentEvent(data="hello\nworld")
+        finally:
+            killed = True
+
+    with httpx.Client(
+        transport=httpx.WSGITransport(
+            SendEventResponse(send_events(), ping_interval=0.1)
+        ),
+        base_url="http://testServer/",
+    ) as client:
+        with client.stream("GET", "/") as resp:
+            resp.raise_for_status()
+            resp.read()
+
+    assert killed
+
+
 @pytest.mark.parametrize(
     "response_class",
     [
@@ -467,34 +592,6 @@ def test_responses_inherit(response_class):
 # ######################################################################################
 
 
-def test_request_response():
-    @request_response
-    def view(request: Request) -> Response:
-        return PlainTextResponse(request.body)
-
-    with httpx.Client(app=view, base_url="http://testServer/") as client:
-        assert client.get("/").text == ""
-        assert client.post("/", content="hello").text == "hello"
-
-
-def test_middleware():
-    @middleware
-    def middleware_func(
-        request: Request, handler: Callable[[Request], Response]
-    ) -> Response:
-        response = handler(request)
-        response.headers["X-Middleware"] = "1"
-        return response
-
-    @request_response
-    @middleware_func
-    def view(request: Request) -> Response:
-        return PlainTextResponse(request.body)
-
-    with httpx.Client(app=view, base_url="http://testServer/") as client:
-        assert client.get("/").headers["X-Middleware"] == "1"
-
-
 def test_router():
     @request_response
     def path(request: Request) -> Response:
@@ -509,7 +606,9 @@ def test_router():
         ("/redirect", redirect),
         ("/{path}", path),
     )
-    with httpx.Client(app=router, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(router)
+    ) as client:
         assert client.get("/").text == "homepage"
         assert client.get("/baize").json() == {"path": "baize"}
         assert client.get("/baize/").status_code == 404
@@ -526,9 +625,11 @@ def test_subpaths():
         return PlainTextResponse(request.get("PATH_INFO", ""))
 
     with httpx.Client(
-        app=Subpaths(
-            ("/frist", root),
-            ("/latest", path),
+        transport=httpx.WSGITransport(
+            Subpaths(
+                ("/frist", root),
+                ("/latest", path),
+            )
         ),
         base_url="http://testServer/",
     ) as client:
@@ -537,9 +638,11 @@ def test_subpaths():
         assert client.get("/latest").text == ""
 
     with httpx.Client(
-        app=Subpaths(
-            ("", path),
-            ("/root", root),
+        transport=httpx.WSGITransport(
+            Subpaths(
+                ("", path),
+                ("/root", root),
+            )
         ),
         base_url="http://testServer/",
     ) as client:
@@ -549,9 +652,11 @@ def test_subpaths():
 
 def test_hosts():
     with httpx.Client(
-        app=Hosts(
-            ("testServer", PlainTextResponse("testServer")),
-            (".*", PlainTextResponse("default host")),
+        transport=httpx.WSGITransport(
+            Hosts(
+                ("testServer", PlainTextResponse("testServer")),
+                (".*", PlainTextResponse("default host")),
+            )
         ),
         base_url="http://testServer/",
     ) as client:
@@ -565,10 +670,13 @@ def test_hosts():
     [
         Files(Path(__file__).absolute().parent.parent / "baize"),
         Files(".", "baize"),
+        Files(".", "baize", handle_404=PlainTextResponse("", 404)),
     ],
 )
 def test_files(app):
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         resp = client.get("/py.typed")
         assert resp.text == ""
 
@@ -603,11 +711,15 @@ def test_files(app):
             )
         ).status_code == 304
 
-        with pytest.raises(HTTPException):
-            client.get("/")
+        if app.handle_404 is None:
+            with pytest.raises(HTTPException):
+                client.get("/")
 
-        with pytest.raises(HTTPException):
-            client.get("/%2E%2E/baize/%2E%2E/%2E%2E/README.md")
+            with pytest.raises(HTTPException):
+                client.get("/%2E%2E/baize/%2E%2E/%2E%2E/README.md")
+
+        else:
+            assert client.get("/").status_code == 404
 
 
 def test_pages(tmpdir):
@@ -620,7 +732,9 @@ def test_pages(tmpdir):
     )
 
     app = Pages(tmpdir)
-    with httpx.Client(app=app, base_url="http://testServer/") as client:
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
         resp = client.get("/")
         assert resp.status_code == 200
         assert resp.text == "<html><body>index</body></html>"
@@ -655,3 +769,93 @@ def test_pages(tmpdir):
 
         with pytest.raises(HTTPException):
             client.get("/d")
+
+    app = Pages(tmpdir, handle_404=PlainTextResponse("", 404))
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
+        assert client.get("/d").status_code == 404
+
+
+# ######################################################################################
+# ################################# Shortcut tests #####################################
+# ######################################################################################
+
+
+def test_request_response():
+    @request_response
+    def view(request: Request) -> Response:
+        return PlainTextResponse(request.body)
+
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(view)
+    ) as client:
+        assert client.get("/").text == ""
+        assert client.post("/", content="hello").text == "hello"
+
+
+def test_decorator():
+    @decorator
+    def decorator_func(
+        request: Request, handler: Callable[[Request], Response]
+    ) -> Response:
+        response = handler(request)
+        response.headers["X-Middleware"] = "1"
+        return response
+
+    @request_response
+    @decorator_func
+    def view(request: Request) -> Response:
+        return PlainTextResponse(request.body)
+
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(view)
+    ) as client:
+        assert client.get("/").headers["X-Middleware"] == "1"
+
+
+def test_middleware():
+    @middleware
+    def m(
+        request: NextRequest, handler: Callable[[NextRequest], NextResponse]
+    ) -> Response:
+        response = handler(request)
+        response.headers["X-Middleware"] = "1"
+        return response
+
+    @m
+    @request_response
+    def view(request: Request) -> Response:
+        return PlainTextResponse(request.body)
+
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(view)
+    ) as client:
+        assert client.get("/").headers["X-Middleware"] == "1"
+
+
+def test_next_request():
+    def app(environ, start_response):
+        request = NextRequest(environ, start_response)
+        request["KEY"] = "VALUE"
+        del request["KEY"]
+        response = NextResponse.from_app(PlainTextResponse(""), request)
+        return response(environ, start_response)
+
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(app)
+    ) as client:
+        response = client.get("/")
+        assert response.status_code == 200
+
+    def read_body(environ, start_response):
+        request = NextRequest(environ, start_response)
+        request.body
+        response = NextResponse.from_app(PlainTextResponse(""), request)
+        return response(environ, start_response)
+
+    with httpx.Client(
+        base_url="http://testServer/", transport=httpx.WSGITransport(read_body)
+    ) as client:
+        with pytest.raises(RuntimeError):
+            client.post("/", content="hello")
